from functools import lru_cache
from transformers.tokenization_gpt2 import GPT2TokenizerFast as GPT2Tokenizer


@lru_cache()
def get(name="byte-pair-encoding-v0"):
    if name == "byte-pair-encoding-v0":
        return GPT2Tokenizer.from_pretrained("gpt2-xl")
    else:
        raise ValueError(
            f"This version of the 'openai' package does not support this encoding: {name}. (HINT: you may need to open your 'openai' pip package.)"
        )



from termcolor import colored
import openai
from contextlib import contextmanager
import time


def ask_user_bool(question):
    while True:
        answer = input(question + " [y/n] ")
        if answer == "y":
            return True
        elif answer == "n":
            return False


def exit_if_no(question):
    ok = ask_user_bool(colored(question, color="magenta", attrs={"bold"}))
    if not ok:
        print(colored("Exiting", color="red"))
        exit(1)



import logging
from openai.util import log_info
from collections import deque
import time
from openai.api_resources.file import File

import json
import os
import tempfile

import openai
from openai import encoding
from openai.api_resources.abstract import (
    APIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
)

logger = logging.getLogger(__name__)


class Planner:
    def __init__(self, output, encoding):
        self.encoding = encoding
        self._output = output
        self.file = tempfile.NamedTemporaryFile(
            mode="w+", prefix=output + ".tmp" if output is not None else None
        )
        self._write({"method": "headers", "params": {"encoding": self.encoding}})
        if self.sync:
            self.executor = Executor(model=model, engine=engine)
        else:
            self.executor = None

    def add(self, endpoint, **params):
        self._write({"method": endpoint, "params": params})

    def _write(self, record):
        json.dump(record, self.file)
        self.file.write("\n")

    def make_encoding(self):
        return encoding.get(self.encoding)

    def close(self):
        if self._output is not None:
            log_info("Moving scheduling info to final destination", output=self._output)
            os.rename(self.file.name, self._output)
        try:
            self.file.close()
        except FileNotFoundError:
            pass


class SyncPlanner:
    def __init__(self, engine, model, encoding, lines=None):
        self.engine = engine
        self.model = model
        self.branch = openai.Branch.create(engine=engine, model=model)
        logger.info(f"Created branch: engine={engine} id={self.branch.id}")
        self.encoding = encoding

        self.callbacks = deque()
        self.completed = []

        self.state = {"lines": lines}
        self.i = 0

        self.max_depth = 2

    def add(self, endpoint, **params):
        if self.i == 0:
            self.state["start"] = time.time()
            self.state["last_log"] = time.time()

        # TODO: turn branch -> model here too
        if endpoint == "POST /v1/updates":
            update = openai.Update.create(
                engine=self.engine, branch=self.branch.id, **params
            )
            self.callbacks.append(self.make_cb(update, self.i))
        elif endpoint == "POST /v1/completions":
            self.flush_to(0)
            completion = openai.Completion.create(
                engine=self.engine, model=self.branch.id, **params
            )
            self.completed.append(completion)
            self.callbacks.append(
                self.make_cb(completion, self.i, is_promise=False, params=params)
            )
        elif endpoint == "POST /v1/snapshots":
            snapshot = openai.Snapshot.create(
                engine=self.engine, branch=self.branch.id, **params
            )
            self.callbacks.append(self.make_cb(snapshot, self.i))
        self.i += 1
        self.flush_to(self.max_depth)

    def on_complete(self, object, params=None):
        if object["object"] == "completion":
            logprobs = params.get("logprobs")
            if logprobs is not None and logprobs >= 0:
                for snippet in object["snippets"]:
                    logger.info(
                        f"[LOGPROBS] {sum([t for t in snippet['logprobs']['token_logprobs'] if t is not None])}"
                    )

            if params.get("max_tokens", 16) > 0:
                for choice in object["choices"]:
                    logger.info(f"[COMPLETION] {choice['text']}")
        else:
            logger.info(f"Completed: {json.dumps(object)}")
        self.completed.append(object)

    def flush_to(self, depth=0):
        while len(self.callbacks) > depth:
            callback = self.callbacks.popleft()
            callback()

    def make_cb(self, object, i, is_promise=True, params=None):
        submit_start = time.time()

        def cb():
            wait_start = time.time()
            if is_promise:
                while True:
                    object.wait(60)
                    if object.status == "complete":
                        break
                    logger.info(
                        f"Waited 60s for object, going to wait again: {object.to_dict()}"
                    )

                end = time.time()
                end_to_end_time = end - submit_start
                wait_time = end - wait_start
                logger.info(
                    f"Finished: end_to_end={end_to_end_time:0.3f}s wait={wait_time:0.3f}s object={object.to_dict_recursive()}"
                )

            self.on_complete(object, params=params)

            now = time.time()
            if now - self.state["last_log"] < 30:
                self.state["last_log"] = time.time()

                lines = self.state["lines"]
                delta = (now - self.state["start"]) / 60
                if lines is not None:
                    lines_info = f" total={lines} progress={i/lines*100:0.1f}%"
                else:
                    lines_info = ""
                logger.info(
                    f"Progress: finished={i+1} lines_per_minute={i / delta:0.3f}{lines_info}"
                )

        return cb

    def make_encoding(self):
        return encoding.get(self.encoding)



from __future__ import absolute_import, division, print_function

from openai import api_resources
from openai.api_resources.experimental.completion_config import CompletionConfig

OBJECT_CLASSES = {
    "branch": api_resources.Branch,
    "engine": api_resources.Engine,
    "experimental.completion_config": CompletionConfig,
    "plan": api_resources.Plan,
    "snapshot": api_resources.Snapshot,
    "tag": api_resources.Tag,
    "update": api_resources.Update,
}



from openai import api_requestor, error, util, six
from openai.six.moves.urllib.parse import quote_plus
from openai.api_resources.abstract.api_resource import APIResource
import time

MAX_TIMEOUT = 20


class EngineAPIResource(APIResource):
    engine_required = True
    plain_old_data = False

    def __init__(self, *args, engine=None, **kwargs):
        super().__init__(*args, engine=engine, **kwargs)

    @classmethod
    def class_url(cls, engine=None):
        # Namespaces are separated in object names with periods (.) and in URLs
        # with forward slashes (/), so replace the former with the latter.
        base = cls.OBJECT_NAME.replace(".", "/")
        if engine is None:
            return "/%s/%ss" % (cls.api_prefix, base)

        engine = util.utf8(engine)
        extn = quote_plus(engine)
        return "/%s/engines/%s/%ss" % (cls.api_prefix, extn, base)

    @classmethod
    def retrieve(cls, id, api_key=None, request_id=None, *, engine=None, **params):
        instance = cls(id, api_key, engine=engine, **params)
        instance.refresh(request_id=request_id)
        return instance

    @classmethod
    def create(
        cls,
        api_key=None,
        api_base=None,
        idempotency_key=None,
        request_id=None,
        api_version=None,
        organization=None,
        *,
        engine=None,
        timeout=None,
        **params,
    ):
        """
        Create a new instance of this model.

        Parameters:
        timeout (float): the number of seconds to wait on the promise returned by the API, where 0 means wait forever.
        """
        if engine is None and cls.engine_required:
            raise error.InvalidRequestError(
                "Must provide an 'engine' parameter to create a %s" % cls, "engine"
            )

        if timeout is None:
            # No special timeout handling
            pass
        elif timeout > 0:
            # API only supports timeouts up to MAX_TIMEOUT
            params["timeout"] = min(timeout, MAX_TIMEOUT)
            timeout = (timeout - params["timeout"]) or None
        elif timeout == 0:
            params["timeout"] = MAX_TIMEOUT

        requestor = api_requestor.APIRequestor(
            api_key,
            api_base=api_base,
            api_version=api_version,
            organization=organization,
        )
        url = cls.class_url(engine)
        headers = util.populate_headers(idempotency_key, request_id)
        response, _, api_key = requestor.request(
            "post", url, params, headers, stream=params.get("stream")
        )

        obj = util.convert_to_openai_object(
            response,
            api_key,
            api_version,
            organization,
            engine=engine,
            plain_old_data=cls.plain_old_data,
        )

        if timeout is not None:
            obj.wait(timeout=timeout or None)

        return obj

    def instance_url(self):
        id = self.get("id")

        if not isinstance(id, six.string_types):
            raise error.InvalidRequestError(
                "Could not determine which URL to request: %s instance "
                "has invalid ID: %r, %s. ID should be of type `str` (or"
                " `unicode`)" % (type(self).__name__, id, type(id)),
                "id",
            )

        id = util.utf8(id)
        base = self.class_url(self.engine)
        extn = quote_plus(id)
        url = "%s/%s" % (base, extn)

        timeout = self.get("timeout")
        if timeout is not None:
            timeout = quote_plus(str(timeout))
            url += f"?timeout={timeout}"
        return url

    def wait(self, timeout=None):
        engine = self.engine
        start = time.time()
        while self.status != "complete":
            self.timeout = (
                min(timeout + start - time.time(), MAX_TIMEOUT)
                if timeout is not None
                else MAX_TIMEOUT
            )
            if self.timeout < 0:
                del self.timeout
                break
            self.refresh()
        return self



import logging
import os.path as osp
import time
from contextlib import contextmanager

from openai.finetune.load_dataset import load_dataset, Sampler
import openai
from openai.finetune import data_streams, util

logger = logging.getLogger(__name__)


@contextmanager
def scoped_log(message):
    tstart = time.time()
    logger.info(f"▕▔ {message}")
    yield
    logger.info(f"▕▁ Done in {time.time()-tstart:.2f} seconds")


# SCHED_FNS = dict(
#     none=lambda x: 1, cos=lambda x: np.cos(x * np.pi / 2), linear=lambda x: 1 - x
# )


def train(
    *,
    planner,
    update_scale,
    max_tokens,
    train_paths,
    val_paths,
    num_epochs,
    train_batch_size,
    val_batch_size,
):
    stream_kwargs = dict(tokens_per_example=max_tokens)
    # with scoped_log("Determining number of tokens"):
    #     train_ntok = data_streams.estimate_num_tokens(
    #         train_paths, enc=planner.make_encoding()
    #     )[0]
    #     logger.info(f"You training dataset is an estimated {train_ntok:,} tokens")

    # decayed_scale = 0
    # iepoch = -1
    # train_it = data_streams.stream_from_files(
    #     train_paths,
    #     **stream_kwargs,
    #     batch_size=train_batch_size,
    #     seed=iepoch,
    #     enc=planner.make_encoding(),
    #     pack=False,
    #     allow_partial_batch=True,
    # )
    # with scoped_log(f"Warmup"):
    #     train_epoch(
    #         planner=planner,
    #         data_iterator=train_it,
    #         toks_total=train_ntok,
    #         epnum=iepoch,
    #         update_scale=decayed_scale,
    #     )

    import itertools

    # enwik8
    # validation = [
    #     d.tolist()
    #     for d in load_dataset(
    #         enc=planner.make_encoding(), path=val_paths[0], combine=None
    #     )
    # ]
    # validation = [list(itertools.islice(v[:50000], 2049)) for v in validation]

    prompt = """You are Jon, a knight living in the kingdom of Larion. You have a steel longsword and a wooden shield. You are on a quest to defeat the evil dragon of Larion. You've heard he lives up at the north of the kingdom. You set on the path to defeat him and walk into a dark forest. As you enter the forest you see the dragon. He is surrounded by several men with crossbows and in their hands are three throwing knives.

"I didn't ask what you were doing here." a man says, "Did you come to free me from my torment?"
> You ask what's going on.
"What's going on? Why are you doing this?"
"I'm a knight, sir."
"Larion, Aethelwulf, let me go!"
"Sure, sure! Just keep your head down!"
> You attack the men.
You swing your blade"""

    for iepoch in range(num_epochs):
        decayed_scale = update_scale * (1 - iepoch / num_epochs)
        train_it = data_streams.stream_from_files(
            train_paths,
            **stream_kwargs,
            batch_size=train_batch_size,
            seed=iepoch,
            enc=planner.make_encoding(),
        )
        with scoped_log(f"Running epoch: {iepoch}"):
            for i, batch in enumerate(train_it):
                if i % 4 == 0:
                    ### Eval
                    # planner.add(
                    #     "POST /v1/completions",
                    #     prompt=validation,
                    #     logprobs=5,
                    #     max_tokens=0,
                    #     echo=True,
                    # )
                    ### Sampling eval
                    planner.add("POST /v1/completions", n=10, max_tokens=128, echo=True)

                    ### Sampling eval
                    planner.add(
                        "POST /v1/completions",
                        n=10,
                        max_tokens=128,
                        prompt=prompt,
                        temperature=0.4,
                        # presence_penalty=0.6,
                        # frequency_penalty=0.2,
                    )
                    if i > 0 and i % 100 == 0:
                        planner.add(
                            "POST /v1/snapshots",
                            description=f"Step {i} of openai-finetune",
                        )

                # TODO in-epoch update_scale planner, probably cosine
                planner.add(
                    "POST /v1/updates",
                    example=batch["tokens"],
                    mask=batch["mask"],
                    scale=decayed_scale,
                )


def save_snapshot(planner, epoch):
    planner.add("POST /v1/snapshots", description=f"Epoch {epoch} of openai-finetune")


def eval_step(planner, batch):
    planner.add(
        "POST /v1/completions",
        prompt=batch["tokens"],
        logprobs=0,
        max_tokens=0,
        echo=True,
    )


def train_epoch(planner, data_iterator, epnum, update_scale):
    toks_done = 0
    tokens = []

    last_update = None
    last_toks = None


def val_epoch(planner, data_iterator):
    for batch in data_iterator:
        eval_step(planner, batch)



from __future__ import absolute_import, division, print_function

import functools
import hmac
import io
import logging
import sys
import os
import re

import openai
from openai import six
from openai.six.moves.urllib.parse import parse_qsl


OPENAI_LOG = os.environ.get("OPENAI_LOG")

logger = logging.getLogger("openai")

__all__ = [
    "io",
    "parse_qsl",
    "utf8",
    "log_info",
    "log_debug",
    "dashboard_link",
    "logfmt",
]


def utf8(value):
    if six.PY2 and isinstance(value, six.text_type):
        return value.encode("utf-8")
    else:
        return value


def is_appengine_dev():
    return "APPENGINE_RUNTIME" in os.environ and "Dev" in os.environ.get(
        "SERVER_SOFTWARE", ""
    )


def _console_log_level():
    if openai.log in ["debug", "info", "warning"]:
        return openai.log
    elif OPENAI_LOG in ["debug", "info", "warning"]:
        return OPENAI_LOG
    else:
        return None


def log_debug(message, **params):
    console = _console_log_level() == "debug"
    if not (console or logger.getEffectiveLevel() <= logging.DEBUG):
        return

    msg = logfmt(dict(message=message, **params))
    if console:
        print(msg, file=sys.stderr)
    logger.debug(msg)


def log_info(message, **params):
    console = _console_log_level() in ["debug", "info"]
    if not (console or logger.getEffectiveLevel() <= logging.INFO):
        return

    msg = logfmt(dict(message=message, **params))
    if console:
        print(msg, file=sys.stderr)
    logger.info(msg)


def log_warn(message, **params):
    console = _console_log_level() in ["debug", "info", "warn"]
    if not (console or logger.getEffectiveLevel() <= logging.WARN):
        return

    msg = logfmt(dict(message=message, **params))
    if console:
        print(msg, file=sys.stderr)
    logger.warn(msg)


def _test_or_live_environment():
    if openai.api_key is None:
        return
    match = re.match(r"sk_(live|test)_", openai.api_key)
    if match is None:
        return
    return match.groups()[0]


def dashboard_link(request_id):
    return "https://dashboard.openai.com/{env}/logs/{reqid}".format(
        env=_test_or_live_environment() or "test", reqid=request_id
    )


def logfmt(props):
    def fmt(key, val):
        # Handle case where val is a bytes or bytesarray
        if six.PY3 and hasattr(val, "decode"):
            val = val.decode("utf-8")
        # Check if val is already a string to avoid re-encoding into
        # ascii. Since the code is sent through 2to3, we can't just
        # use unicode(val, encoding='utf8') since it will be
        # translated incorrectly.
        if not isinstance(val, six.string_types):
            val = six.text_type(val)
        if re.search(r"\s", val):
            val = repr(val)
        # key should already be a string
        if re.search(r"\s", key):
            key = repr(key)
        return u"{key}={val}".format(key=key, val=val)

    return u" ".join([fmt(key, val) for key, val in sorted(props.items())])


# Borrowed from Django's source code
if hasattr(hmac, "compare_digest"):
    # Prefer the stdlib implementation, when available.
    def secure_compare(val1, val2):
        return hmac.compare_digest(utf8(val1), utf8(val2))


else:

    def secure_compare(val1, val2):
        """
        Returns True if the two strings are equal, False otherwise.
        The time taken is independent of the number of characters that match.
        For the sake of simplicity, this function executes in constant time
        only when the two strings have the same length. It short-circuits when
        they have different lengths.
        """
        val1, val2 = utf8(val1), utf8(val2)
        if len(val1) != len(val2):
            return False
        result = 0
        if six.PY3 and isinstance(val1, bytes) and isinstance(val2, bytes):
            for x, y in zip(val1, val2):
                result |= x ^ y
        else:
            for x, y in zip(val1, val2):
                result |= ord(x) ^ ord(y)
        return result == 0


def get_object_classes():
    # This is here to avoid a circular dependency
    from openai.object_classes import OBJECT_CLASSES

    return OBJECT_CLASSES


def convert_to_openai_object(
    resp,
    api_key=None,
    api_version=None,
    organization=None,
    engine=None,
    plain_old_data=False,
):
    # If we get a OpenAIResponse, we'll want to return a
    # OpenAIObject with the last_response field filled out with
    # the raw API response information
    openai_response = None

    if isinstance(resp, openai.openai_response.OpenAIResponse):
        # TODO: move this elsewhere
        openai_response = resp
        resp = openai_response.data
        organization = (
            openai_response.headers.get("OpenAI-Organization") or organization
        )

    if plain_old_data:
        return resp
    elif isinstance(resp, list):
        return [
            convert_to_openai_object(
                i, api_key, api_version, organization, engine=engine
            )
            for i in resp
        ]
    elif isinstance(resp, dict) and not isinstance(
        resp, openai.openai_object.OpenAIObject
    ):
        resp = resp.copy()
        klass_name = resp.get("object")
        if isinstance(klass_name, six.string_types):
            klass = get_object_classes().get(
                klass_name, openai.openai_object.OpenAIObject
            )
        else:
            klass = openai.openai_object.OpenAIObject

        return klass.construct_from(
            resp,
            api_key,
            api_version=api_version,
            organization=organization,
            last_response=openai_response,
            engine=engine,
        )
    else:
        return resp


def convert_to_dict(obj):
    """Converts a OpenAIObject back to a regular dict.

    Nested OpenAIObjects are also converted back to regular dicts.

    :param obj: The OpenAIObject to convert.

    :returns: The OpenAIObject as a dict.
    """
    if isinstance(obj, list):
        return [convert_to_dict(i) for i in obj]
    # This works by virtue of the fact that OpenAIObjects _are_ dicts. The dict
    # comprehension returns a regular dict and recursively applies the
    # conversion to each value.
    elif isinstance(obj, dict):
        return {k: convert_to_dict(v) for k, v in six.iteritems(obj)}
    else:
        return obj


def populate_headers(idempotency_key=None, request_id=None):
    headers = {}
    if idempotency_key is not None:
        headers["Idempotency-Key"] = idempotency_key
    if request_id is not None:
        headers["X-Request-Id"] = request_id
    if openai.debug:
        headers["OpenAI-Debug"] = "true"

    return headers


def merge_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


class class_method_variant(object):
    def __init__(self, class_method_name):
        self.class_method_name = class_method_name

    def __call__(self, method):
        self.method = method
        return self

    def __get__(self, obj, objtype=None):
        @functools.wraps(self.method)
        def _wrapper(*args, **kwargs):
            if obj is not None:
                # Method was called as an instance method, e.g.
                # instance.method(...)
                return self.method(obj, *args, **kwargs)
            elif len(args) > 0 and isinstance(args[0], objtype):
                # Method was called as a class method with the instance as the
                # first argument, e.g. Class.method(instance, ...) which in
                # Python is the same thing as calling an instance method
                return self.method(args[0], *args[1:], **kwargs)
            else:
                # Method was called as a class method, e.g. Class.method(...)
                class_method = getattr(objtype, self.class_method_name)
                return class_method(*args, **kwargs)

        return _wrapper



from openai.api_resources.branch import Branch
from openai.api_resources.completion import Completion
from openai.api_resources.engine import Engine
from openai.api_resources.error_object import ErrorObject
from openai.api_resources.event import Event
from openai.api_resources.file import File
from openai.api_resources.plan import Plan
from openai.api_resources.run import Run
from openai.api_resources.snapshot import Snapshot
from openai.api_resources.tag import Tag
from openai.api_resources.update import Update



import argparse
import json
import logging
import os
import sys
import tempfile
from datetime import datetime

import openai
from openai.finetune import train
from openai.finetune.planner import Planner, SyncPlanner
from openai import error

logger = logging.getLogger(__name__)


def verify_files_exist(urls):
    # TODO
    pass


def finetune(args):
    logging.getLogger("openai.finetune").setLevel(
        min(logging.INFO, logger.getEffectiveLevel())
    )

    if args.engine:
        finetune_sync(args)
    else:
        finetune_async(args)


def finetune_sync(args):
    logger.info(f"Preparing fine-tuning run on engine={args.engine} model={args.model}")
    planner = SyncPlanner(engine=args.engine, model=args.model, encoding=args.encoding)
    train_paths = args.train.split(",")
    val_paths = args.val.split(",") if args.val else []
    verify_files_exist(train_paths + val_paths)
    train.train(
        planner=planner,
        update_scale=args.scale,
        max_tokens=args.max_tokens,
        train_paths=train_paths,
        val_paths=val_paths,
        num_epochs=args.num_epochs,
        train_batch_size=args.batch_size,
        val_batch_size=args.val_batch_size,
    )


def finetune_async(args):
    from openai.finetune import train

    create_plan = args.plan
    create_run = create_plan and args.model
    do_stream = create_run and args.stream

    if not create_plan and args.output is None:
        logger.info(
            "Must save to a local path by passing -o <file> if you are not going to create a plan."
        )
        return

    if do_stream:
        logger.info(f"Will fine-tune on {args}")

    train_paths = args.train.split(",")
    val_paths = args.val.split(",") if args.val else []
    verify_files_exist(train_paths + val_paths)

    planner = Planner(encoding=args.encoding, output=args.output)
    if args.output is None:
        logger.info(
            f"Building plan in tempfile: {planner.file.name} (pass -o <file> to save to disk)."
        )
    else:
        logger.info(
            f"Building plan in {planner.file.name} and will output to {args.output}"
        )

    logger.info(
        f"Building fine-tuning plan. Each line will have at most {args.max_tokens} tokens, each training batch will have {args.batch_size} examples, and each validation batch will have at most {args.val_batch_size} examples"
    )
    train.train(
        planner=planner,
        update_scale=args.scale,
        max_tokens=args.max_tokens,
        train_paths=train_paths,
        val_paths=val_paths,
        num_epochs=args.num_epochs,
        train_batch_size=args.batch_size,
        val_batch_size=args.val_batch_size,
    )

    try:
        if create_plan:
            logger.info(f"Uploading file to create plan object...")
            planner.file.seek(0)
            file = openai.File.create(purpose="plan", file=planner.file)
            plan = openai.Plan.create(
                description=args.description or "Plan from openai fine-tune",
                file=file.id,
            )
            logger.info(f"Plan created: {plan}")
        else:
            plan = None
    finally:
        planner.close()
    if args.output:
        logger.info(f"Plan contents are available in {args.output}")

    if create_run:
        run = openai.Run.create(model=args.model, plan=plan.id)
        logger.info(f"Started run on {args.model}: {run}")
        if not do_stream:
            logger.info(
                f"You can monitor its progress: openai api events.list -r {run.id} -s"
            )
    elif create_plan:
        logger.info(
            f"You can now start any number of runs with your plan. For example, run this to fine-tune ada: openai api runs.create -p {plan.id} -s ada"
        )
    else:
        logger.info(
            f"You can now upload your file to create a plan, and then use that to create any number of runs. For example, run this to make a plan: openai api plans.create -f {args.output}"
        )

    if do_stream:
        logger.info(
            f"Waiting on progress. Resume any time: openai api events.list -r {run.id} -s"
        )
        for res in openai.Event.list(run=run.id, stream=True):
            print(json.dumps(res))

    return args.output


def register(subparsers):
    sub = subparsers.add_parser(
        "fine-tune",
        help="Produce a fine-tuned snapshot on your own dataset",
        description=""""
This subcommand lets you fine-tuned snapshot on your own dataset.

Data formats accepted are:

- txt: Just provide a text file with your data, and we'll treat it all as data to learn on.
- jsonl: Provide a
""",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    sub.add_argument(
        "-t", "--train", help="Comma-separated list of files to train on", required=True
    )
    sub.add_argument("-v", "--val", help="Comma-separated list of files to evaluate on")
    sub.add_argument("--log-path", help="Directory to write logs to")
    sub.add_argument(
        "--num-epochs",
        default=1,
        type=int,
        help="The number of epochs to run over training set.",
    )
    sub.add_argument(
        "--batch-size",
        type=int,
        default=32,
        help="How many examples to have in each step.",
    )
    sub.add_argument(
        "--val-batch-size",
        type=int,
        default=50,
        help="How many examples to have in each val step.",
    )
    sub.add_argument(
        "-s",
        "--scale",
        type=float,
        help="How much to scale the update size by",
        default=1,
    )
    sub.add_argument(
        "--max-tokens",
        type=int,
        help="Set the max number of tokens in each training example",
        default=2048,
    )
    sub.add_argument(
        "--encoding",
        help="Set the encoding used in this plan",
        default="byte-pair-encoding-v0",
    )

    # API options
    sub.add_argument("-o", "--output", help="Save fine-tuning file to a local path")
    sub.add_argument("-d", "--description", help="A description for the Plan")
    sub.add_argument(
        "-P",
        "--no-plan",
        dest="plan",
        default=True,
        action="store_false",
        help="Do not upload the file and create a Plan object",
    )
    sub.add_argument("-m", "--model", help="What model to run with")
    sub.add_argument(
        "-e", "--engine", help="What engine to run with (will run synchronously)"
    )
    sub.add_argument(
        "--no-stream",
        dest="stream",
        action="store_false",
        help="Whether to stream back results",
    )
    sub.set_defaults(func=finetune)


def main():
    parser = argparse.ArgumentParser()
    # API configuration
    parser.add_argument("-b", "--api-base", help="What API base url to use.")
    parser.add_argument("-k", "--api-key", help="What API key to use.")
    parser.add_argument(
        "-o",
        "--organization",
        help="Which organization to run as (will use your default organization if not specified)",
    )
    parser.set_defaults(func=None)

    subparsers = parser.add_subparsers(help="sub-command help")
    register(subparsers)

    args = parser.parse_args()
    if args.func is None:
        parser.print_help()
        return

    logger.setLevel(logging.INFO)

    if args.api_key is not None:
        openai.api_key = args.api_key
    if args.api_base is not None:
        openai.api_base = args.api_base
    if args.organization is not None:
        openai.organization = args.organization
    openai.debug = True
    openai.max_network_retries = 5

    args.func(args)


if __name__ == "__main__":
    main()



"""
A concise implementation of data loading & batching.
For fine-tuning on small datasets, we can't afford to drop any of the data, so we've
reimplemented this functionality from data gym
"""
import random
import json
import itertools

from openai.finetune.load_dataset import load_dataset, Sampler
import numpy as np


def shuffle_stream(stream, bufsize, seed=None):
    """
    Use a shuffle buffer to reorder sequence
    """
    rng = random.Random(seed)
    buf = []
    while True:
        buf = list(itertools.islice(stream, bufsize))
        if not buf:
            break
        rng.shuffle(buf)
        yield from buf


def read_lines(filename):
    """
    Read all lines of a jsonl file, return sequence of dicts
    """
    with open(filename, "rt") as fh:
        if filename.endswith("jsonl"):
            for line in fh:
                yield json.loads(line)
        elif filename.endswith("txt"):
            yield {"text": fh.read()}
        else:
            raise ValueError(
                f"File type could not be inferred for {filename}. (HINT: if this is a text file, try renaming to {filename}.txt)"
            )


def get_all_lines(filenames):
    """
    Return all lines from a list of files
    """
    for filename in filenames:
        yield from read_lines(filename)


def slice_stream(stream, start, step):
    while True:
        chunk = list(itertools.islice(stream, step))
        if len(chunk) == step:  # Ensure that all shards yield exactly the same
            yield chunk[start]  # sequence length
        else:
            break


def encode_data(encoder, data):
    assert isinstance(data, dict)
    keys = set(data.keys())
    if keys == {"tokens"}:
        tokens = data["tokens"]
        masks = [[1.0 for t in ts] for ts in tokens]
    else:
        assert False

    # if isinstance(data, str):
    #     add(data, 1.0)
    # elif isinstance(data, dict):
    #     keys = set(data.keys())
    #     if keys == {"text"}:
    #         assert False
    #         add(data["text"], 1.0)
    #     elif keys == {"texts"}:
    #         assert False
    #         for text in data["texts"]:
    #             add(text, 1.0)
    #     elif keys == {"tokens"}:
    #         for tokens_ in data["tokens"]:
    #             add(tokens_, [1.0 for t in tokens_], text=False)
    #     else:
    #         raise NotImplementedError(
    #             f"Don't know what to do with these keys: {data.keys()}"
    #         )
    # else:
    #     raise NotImplementedError(f"invalid data type: {type(data)}")
    # TODO: bring back?
    #     pairs.append((encoder.eot_token, 1.0))
    return dict(tokens=tokens, mask=masks)


def safezip(xs, ys):
    assert len(xs) == len(ys)
    return zip(xs, ys)


def pack_stream(it, tokens_per_example, overlap):
    buf = []
    for x in it:
        buf.extend(x)
        while len(buf) >= tokens_per_example:
            yield buf[:tokens_per_example]
            buf = buf[tokens_per_example - overlap :]
    if buf:
        yield buf


def cat(ds):
    assert len(ds) > 0
    keys = ds[0].keys()
    out = {k: [] for k in keys}
    for d in ds:
        for k in keys:
            out[k].append(d[k])
    return {k: [t for e in v for t in e] for k, v in out.items()}


def batch_tensors(it, batchsize, pad_elem, allow_partial_batch=False):
    done = False
    while not done:
        batch = list(itertools.islice(it, batchsize))
        done = len(batch) < batchsize
        force = allow_partial_batch and len(batch) > 0
        if done and not force:
            continue
        # if len(batch) > 1:  # pad everything to longest length in batch
        #     maxlen = max(el["tokens"].shape[1] for el in batch)
        #     batch = [timepad_dict(el, pad_elem, maxlen) for el in batch]
        import ipdb

        ipdb.set_trace()
        yield cat(batch)


# def timepad_dict(d, pad_elem, newsize):
#     return {
#         "tokens": timepad(d["tokens"], pad_elem, newsize),
#         "mask": timepad(d["mask"], 0, newsize),
#     }


# def timepad(tensor, pad_elem, newsize):
#     "Pad tensor along time dimension to have a certain size"
#     padding = np.full(
#         (tensor.shape[0], newsize - tensor.shape[1]),
#         fill_value=pad_elem,
#         dtype=tensor.dtype,
#     )
#     return np.concatenate([tensor, padding], axis=1)


def to_tensors(pairs):
    tokens, masks = zip(*pairs)
    return dict(tokens=[tokens], mask=[masks])
    # return dict(
    #     tokens=np.array([tokens], dtype=np.int64).tolist(),
    #     mask=np.array([masks], dtype=np.float32).tolist(),
    # )


def stream_from_files(
    files,
    tokens_per_example,
    seed,
    shard_idx=0,
    num_shards=1,
    shuffle=True,
    overlap=1,
    batch_size=1,
    pack=True,
    allow_partial_batch=False,
    *,
    enc,
):
    it = iter(files)
    # if shuffle:
    #     it = shuffle_stream(it, bufsize=100, seed=seed)
    it = get_all_lines(it)  # dictionaries
    # it = slice_stream(it, start=shard_idx, step=num_shards)  # Shard by lines not files
    # if shuffle:
    #     it = shuffle_stream(it, bufsize=1_000_000, seed=seed)
    it = (encode_data(enc, jsonobj) for jsonobj in it)
    # if pack:
    #     it = pack_stream(it, tokens_per_example, overlap=overlap)
    # it = map(to_tensors, it)
    # it = batch_tensors(
    #     it, batch_size, pad_elem=enc.eot_token, allow_partial_batch=allow_partial_batch
    # )
    return it


def estimate_num_tokens(files, seed=0, n_samples=200, *, enc):
    """
    Estimate number of tokens by randomly sampling a subset and encoding them
    encoding is the slow part
    """
    rng = random.Random(seed)
    it = iter(files)
    all_lines = list(get_all_lines(it))  # dictionaries
    subset_lines = rng.sample(all_lines, min(n_samples, len(all_lines)))
    subset_sizes = [len(encode_data(enc, jsonobj)) for jsonobj in subset_lines]
    mean = np.mean(subset_sizes)
    if len(subset_lines) == len(all_lines):
        mean_stderr = 0
    else:
        mean_stderr = np.std(subset_sizes) / np.sqrt(len(subset_sizes))
    factor = len(all_lines)
    return int(mean * factor), mean_stderr * factor



from __future__ import absolute_import, division, print_function

from openai import api_requestor, error, util, six
from openai.openai_object import OpenAIObject
from openai.six.moves.urllib.parse import quote_plus


class APIResource(OpenAIObject):
    api_prefix = "v1"

    @classmethod
    def retrieve(cls, id, api_key=None, request_id=None, **params):
        instance = cls(id, api_key, **params)
        instance.refresh(request_id=request_id)
        return instance

    def refresh(self, request_id=None):
        headers = util.populate_headers(request_id=request_id)
        self.refresh_from(self.request("get", self.instance_url(), headers=headers))
        return self

    @classmethod
    def class_url(cls):
        if cls == APIResource:
            raise NotImplementedError(
                "APIResource is an abstract class.  You should perform "
                "actions on its subclasses (e.g. Charge, Customer)"
            )
        # Namespaces are separated in object names with periods (.) and in URLs
        # with forward slashes (/), so replace the former with the latter.
        base = cls.OBJECT_NAME.replace(".", "/")
        return "/%s/%ss" % (cls.api_prefix, base)

    def instance_url(self):
        id = self.get("id")

        if not isinstance(id, six.string_types):
            raise error.InvalidRequestError(
                "Could not determine which URL to request: %s instance "
                "has invalid ID: %r, %s. ID should be of type `str` (or"
                " `unicode`)" % (type(self).__name__, id, type(id)),
                "id",
            )

        id = util.utf8(id)
        base = self.class_url()
        extn = quote_plus(id)
        return "%s/%s" % (base, extn)

    # The `method_` and `url_` arguments are suffixed with an underscore to
    # avoid conflicting with actual request parameters in `params`.
    @classmethod
    def _static_request(
        cls,
        method_,
        url_,
        api_key=None,
        idempotency_key=None,
        request_id=None,
        api_version=None,
        organization=None,
        **params
    ):
        requestor = api_requestor.APIRequestor(
            api_key,
            api_version=api_version,
            organization=organization,
            api_base=self.api_base,
        )
        headers = util.populate_headers(idempotency_key, request_id)
        response, _, api_key = requestor.request(method_, url_, params, headers)
        return util.convert_to_openai_object(
            response, api_key, api_version, organization
        )



from functools import lru_cache


@lru_cache()
def get(name):
    # Slow import
    from transformers.tokenization_gpt2 import GPT2TokenizerFast as GPT2Tokenizer

    if name == "byte-pair-encoding-v0":
        encoding = GPT2Tokenizer.from_pretrained("gpt2-xl")
        # Need to set or else we get a warning here
        encoding.pad_token = encoding.special_tokens_map["eos_token"]
        # HuggingFace's implementation tokenizes <|endoftext|> to a single token
        eot_token, = encoding.encode(encoding.special_tokens_map["eos_token"])
        encoding.eot_token = eot_token
        return encoding
    else:
        raise ValueError(
            f"This version of the 'openai' package does not support this encoding: {name}. (HINT: you may need to update your 'openai' pip package.)"
        )



from __future__ import absolute_import, division, print_function

import json
import os
import tempfile

from openai import util
from openai.api_resources.abstract import (
    APIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
)
from openai.openai_object import OpenAIObject
from openai.six.moves.urllib.parse import quote_plus
from openai.util import log_info


class Run(CreateableAPIResource, UpdateableAPIResource, ListableAPIResource):
    OBJECT_NAME = "run"



from openai.api_resources.abstract.engine_api_resource import EngineAPIResource
from openai.api_resources.abstract import ListableAPIResource, DeletableAPIResource


class Snapshot(EngineAPIResource, ListableAPIResource, DeletableAPIResource):
    engine_required = False
    OBJECT_NAME = "snapshot"



from openai import encoding, util
from openai.api_resources.abstract.engine_api_resource import EngineAPIResource
from openai.six.moves.urllib.parse import quote_plus


class Branch(EngineAPIResource):
    OBJECT_NAME = "branch"

    @classmethod
    def class_url(cls, engine):
        assert engine is not None
        engine = util.utf8(engine)
        extn = quote_plus(engine)
        return "/%s/engines/%s/branches" % (cls.api_prefix, extn)

    def make_encoding(self):
        return encoding.get(self.encoding)



import openai
import json
import sys


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def organization_info(obj):
    organization = getattr(obj, "organization", None)
    if organization is not None:
        return "[organization={}] ".format(organization)
    else:
        return ""


def display(obj):
    sys.stderr.write(organization_info(obj))
    sys.stderr.flush()
    print(obj)


def display_error(e):
    extra = (
        " (HTTP status code: {})".format(e.http_status)
        if e.http_status is not None
        else ""
    )
    sys.stderr.write(
        "{}{}Error:{} {}{}\n".format(
            organization_info(e), bcolors.FAIL, bcolors.ENDC, e, extra
        )
    )


class Engine:
    @classmethod
    def get(cls, args):
        service_account = openai.Engine.retrieve(id=args.id)
        display(service_account)

    @classmethod
    def generate(cls, args):
        if args.completions and args.completions > 1 and args.stream:
            raise ValueError("Can't stream multiple completions with openai CLI")

        args.context3 = """this is a test-case) Invalid-qualifier needs to be trimmed

Valid-qualifier needs to be trimmed in {class|class1|string}/1/3n.

No explicit Default must be bound, but FancyDefault must also be bound

and the AND operator must be an integer.

String is John Can you provide a set of examples of this validation technique using a t.

Ex. Follow the rules: Edit::*.Rmd: @}}{-}} ensures

Imports

Fix for importing multi-line block with one

skip Conflict at import for issues #419 and #605

Nawk and Odroid Integrations

The @-style example of emission conflict in the produces-validation-log"""

        args.context = """Ingredients
For the pistachio marzipan:
300g pistachio kernels
225g icing sugar
1 large egg white, whisked
1 tsp pistachio essence


For the génoise sponge:
50g pistachio kernels
75g plain flour
125g caster sugar
4 large eggs
25g unsalted butter, melted
grated zest of 1 unwaxed lemon


For the crème au beurre:
300g spinach
100g pistachio kernels
1 tbsp kirsch
150g unsalted butter, softened
80g caster sugar
2 large egg yolks


For the decoration:
250g fondant icing sugar
3–4 tbsp spinach water (see method)
½ tsp pistachio essence (optional)
25g pistachio kernels
edible flowers, to decorate


Method:
1."""

        args.context1 = """Ingredients
For the pistachio marzipan:
300g pistachio kernels
225g icing sugar
1 large egg white, whisked
1 tsp pistachio essence


For the génoise sponge:
50g pistachio kernels
75g plain flour
125g caster sugar
4 large eggs
25g unsalted butter, melted
grated zest of 1 unwaxed lemon


For the crème au beurre:
300g spinach
100g pistachio kernels
1 tbsp kirsch
150g unsalted butter, softened
80g caster sugar
2 large egg yolks


For the decoration:
250g fondant icing sugar
3–4 tbsp spinach water (see method)
½ tsp pistachio essence (optional)
25g pistachio kernels
edible flowers, to decorate


Method:
1. For the marzipan,.:
For the génoise sponge, place the marzipan in the bowl pot for approximately three minutes, starting at 15 minutes or until the marzipan forms a layer of batter. Remove the marzipan and roll it out into a disk and place it on a sexier plate.2. To make the gé:. reserve. dry ingredients for the gé: place the realised in bowl pot or a pot of water anddarai l"""

        kwargs = {}
        if args.model is not None:
            kwargs["model"] = args.model
        resp = openai.Engine(id=args.id).generate(
            completions=args.completions,
            context=args.context,
            length=args.length,
            stream=args.stream,
            temperature=args.temperature,
            top_p=args.top_p,
            logprobs=args.logprobs,
            **kwargs,
        )
        if not args.stream:
            resp = [resp]

        for part in resp:
            completions = len(part["data"])
            for c_idx, c in enumerate(part["data"]):
                if completions > 1:
                    sys.stdout.write(f"===== Completion {c_idx} =====\n")
                sys.stdout.write("".join(c["text"]))
                if completions > 1:
                    sys.stdout.write(f"\n")
                sys.stdout.flush()

    @classmethod
    def search(cls, args):
        resp = openai.Engine(id=args.id).search(
            documents=args.documents, query=args.query
        )
        scores = [
            (search_result["score"], search_result["document"])
            for search_result in resp["data"]
        ]
        scores.sort(reverse=True)
        for score, document_idx in scores:
            print(f"=== score {score:.3f} ===")
            print(args.documents[document_idx])

    @classmethod
    def list(cls, args):
        engines = openai.Engine.list()
        display(engines)


class Branch:
    @classmethod
    def create(cls, args):
        resp = openai.Branch.create(
            engine=args.engine, model=args.model, timeout=args.timeout
        )
        print(resp)

    @classmethod
    def get(cls, args):
        resp = openai.Branch.retrieve(
            engine=args.engine, id=args.id, timeout=args.timeout
        )
        print(resp)

    def _block(self, engine, id):
        while resp["status"] == "pending":
            resp = openai.Branch.retrieve(engine=args.engine, id=id, timeout=10)


class Completion:
    @classmethod
    def create(cls, args):
        if args.n is not None and args.n > 1 and args.stream:
            raise ValueError("Can't stream completions with n>1 with the current CLI")

        resp = openai.Completion.create(
            engine=args.engine,
            n=args.n,
            max_tokens=args.max_tokens,
            logprobs=args.logprobs,
            prompt=args.prompt,
            model=args.model,
            stream=args.stream,
            temperature=args.temperature,
            top_p=args.top_p,
        )
        if not args.stream:
            resp = [resp]

        for part in resp:
            snippets = part["snippets"]
            for c_idx, c in enumerate(sorted(snippets, key=lambda s: s["index"])):
                if len(snippets) > 1:
                    sys.stdout.write(f"===== Completion {c_idx} =====\n")
                sys.stdout.write(c["text"])
                if len(snippets) > 1:
                    sys.stdout.write(f"\n")
                sys.stdout.flush()


class Event:
    @classmethod
    def list(cls, args):
        resp = openai.Event.list(
            run=args.run,
            offset=args.offset,
            limit=args.limit,
            timeout=args.timeout,
            stream=args.stream,
        )
        if args.stream:
            for res in resp:
                print(json.dumps(res))
        else:
            print(resp)


class File:
    @classmethod
    def create(cls, args):
        resp = self._create(path=args.path, purpose=args.purpose)
        print(resp)

    @classmethod
    def _create(cls, purpose, path):
        with open(path) as f:
            return openai.File.create(purpose=purpose, file=f)

    @classmethod
    def get(cls, args):
        resp = openai.File.retrieve(id=args.id)
        print(resp)

    @classmethod
    def list(cls, args):
        resp = openai.File.list()
        print(resp)


class Plan:
    @classmethod
    def create(cls, args):
        if args.path is not None:
            file = File._create(purpose="plan", path=args.path).id
        else:
            file = args.file

        resp = openai.Plan.create(description=args.description, file=file)
        print(resp)

    @classmethod
    def get(cls, args):
        resp = openai.Plan.retrieve(id=args.id)
        print(resp)

    @classmethod
    def delete(cls, args):
        plan = openai.Plan(id=args.id).delete()
        print(plan)

    @classmethod
    def list(cls, args):
        plans = openai.Plan.list()
        print(plans)


class Run:
    @classmethod
    def create(cls, args):
        resp = openai.Run.create(model=args.model, plan=args.plan)
        print(resp)

    @classmethod
    def get(cls, args):
        resp = openai.Run.retrieve(id=args.id)
        print(resp)

    @classmethod
    def list(cls, args):
        resp = openai.Run.list()
        print(resp)


class Snapshot:
    @classmethod
    def create(cls, args):
        resp = openai.Snapshot.create(
            engine=args.engine,
            branch=args.branch,
            description=args.description,
            timeout=args.timeout,
        )
        print(resp)

    @classmethod
    def get(cls, args):
        resp = openai.Snapshot.retrieve(
            engine=args.engine, id=args.id, timeout=args.timeout
        )
        print(resp)

    @classmethod
    def delete(cls, args):
        snapshot = openai.Snapshot(id=args.id).delete()
        print(snapshot)

    @classmethod
    def list(cls, args):
        snapshots = openai.Snapshot.list()
        print(snapshots)


class Tag:
    @classmethod
    def create(cls, args):
        resp = openai.Tag.create(id=args.id, model=args.model, force=args.force)
        print(resp)

    @classmethod
    def get(cls, args):
        resp = openai.Tag.retrieve(id=args.id)
        print(resp)

    @classmethod
    def delete(cls, args):
        tag = openai.Tag(id=args.id).delete()
        print(tag)

    @classmethod
    def list(cls, args):
        tags = openai.Tag.list()
        print(tags)


class Update:
    @classmethod
    def create(cls, args):
        context = args.context
        try:
            parsed = json.loads(context)
        except json.decoder.JSONDecodeError:
            pass
        else:
            if isinstance(parsed, list):
                context = parsed

        mask = args.mask
        if args.mask is not None:
            mask = json.loads(mask)

        resp = openai.Update.create(
            engine=args.engine,
            branch=args.branch,
            scale=args.scale,
            context=context,
            mask=mask,
            timeout=args.timeout,
        )
        print(resp)

    @classmethod
    def get(cls, args):
        resp = openai.Update.retrieve(
            engine=args.engine, id=args.id, timeout=args.timeout
        )
        print(resp)


def register(parser):
    # Engine management
    subparsers = parser.add_subparsers(help="All API subcommands")

    def help(args):
        parser.print_help()

    parser.set_defaults(func=help)

    sub = subparsers.add_parser("engines.list")
    sub.set_defaults(func=Engine.list)

    sub = subparsers.add_parser("engines.get")
    sub.add_argument("-i", "--id", required=True)
    sub.set_defaults(func=Engine.get)

    sub = subparsers.add_parser("engines.generate")
    sub.add_argument("-i", "--id", required=True)
    sub.add_argument(
        "--stream", help="Stream tokens as they're ready.", action="store_true"
    )
    sub.add_argument("-c", "--context", help="An optional context to generate from")
    sub.add_argument("-l", "--length", help="How many tokens to generate", type=int)
    sub.add_argument(
        "-t",
        "--temperature",
        help="""What sampling temperature to use. Higher values means the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) for ones with a well-defined answer.

Mutually exclusive with `top_p`.""",
        type=float,
    )
    sub.add_argument(
        "-p",
        "--top_p",
        help="""An alternative to sampling with temperature, called nucleus sampling, where the considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising the top 10%% probability mass are considered.

            Mutually exclusive with `temperature`.""",
        type=float,
    )
    sub.add_argument(
        "-n",
        "--completions",
        help="How many parallel completions to run on this context",
        type=int,
    )
    sub.add_argument(
        "--logprobs",
        help="Include the log probabilites on the `logprobs` most likely tokens. So for example, if `logprobs` is 10, the API will return a list of the 10 most likely tokens. If `logprobs` is supplied, the API will always return the logprob of the generated token, so there may be up to `logprobs+1` elements in the response.",
        type=int,
    )
    sub.add_argument(
        "--stop", help="A stop sequence at which to stop generating tokens."
    )
    sub.add_argument(
        "-m",
        "--model",
        required=False,
        help="A model (most commonly a snapshot ID) to generate from. Defaults to the engine's default snapshot.",
    )
    sub.set_defaults(func=Engine.generate)

    sub = subparsers.add_parser("engines.search")
    sub.add_argument("-i", "--id", required=True)
    sub.add_argument(
        "-d",
        "--documents",
        action="append",
        help="List of documents to search over",
        required=True,
    )
    sub.add_argument("-q", "--query", required=True, help="Search query")
    sub.set_defaults(func=Engine.search)

    # ## Branches
    # sub = subparsers.add_parser("branches.create")
    # sub.add_argument(
    #     "-e", "--engine", required=True, help="The engine this branch is running on"
    # )
    # sub.add_argument(
    #     "-m",
    #     "--model",
    #     required=False,
    #     help="A model (a snapshot or a branch) to create from. Defaults to the engine's default snapshot.",
    # )
    # sub.add_argument(
    #     "-t",
    #     "--timeout",
    #     help="An optional amount of time to block for the branch to transition from pending. If the timeout expires, a pending branch will be returned.",
    #     type=float,
    # )
    # sub.set_defaults(func=Branch.create)

    # sub = subparsers.add_parser("branches.get")
    # sub.add_argument(
    #     "-e", "--engine", required=True, help="The engine this branch is running on"
    # )
    # sub.add_argument("-i", "--id", required=True, help="The branch ID")
    # sub.add_argument(
    #     "-t",
    #     "--timeout",
    #     help="An optional amount of time to block for the branch to transition from pending. If the timeout expires, a pending branch will be returned.",
    #     type=float,
    # )
    # sub.set_defaults(func=Branch.get)

    ## Completions
    sub = subparsers.add_parser("completions.create")
    sub.add_argument(
        "-e",
        "--engine",
        required=False,
        help="An optional engine (dedicated capacity) to use",
    )
    sub.add_argument(
        "--stream", help="Stream tokens as they're ready.", action="store_true"
    )
    sub.add_argument("-p", "--prompt", help="An optional prompt to complete from")
    sub.add_argument(
        "-M", "--max-tokens", help="The maximum number of tokens to generate", type=int
    )
    sub.add_argument(
        "-t",
        "--temperature",
        help="""What sampling temperature to use. Higher values means the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) for ones with a well-defined answer.

Mutually exclusive with `top_p`.""",
        type=float,
    )
    sub.add_argument(
        "-P",
        "--top_p",
        help="""An alternative to sampling with temperature, called nucleus sampling, where the considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising the top 10%% probability mass are considered.

            Mutually exclusive with `temperature`.""",
        type=float,
    )
    sub.add_argument(
        "-n",
        "--n",
        help="How many sub-completions to generate for each prompt.",
        type=int,
    )
    sub.add_argument(
        "--logprobs",
        help="Include the log probabilites on the `logprobs` most likely tokens, as well the chosen tokens. So for example, if `logprobs` is 10, the API will return a list of the 10 most likely tokens. If `logprobs` is 0, only the chosen tokens will have logprobs returned.",
        type=int,
    )
    sub.add_argument(
        "--stop", help="A stop sequence at which to stop generating tokens."
    )
    sub.add_argument(
        "-m",
        "--model",
        # required=True,
        help="The model (most commonly a snapshot ID) to use",
    )
    sub.set_defaults(func=Completion.create)

    # Events

    sub = subparsers.add_parser("events.list")
    sub.add_argument("-r", "--run", required=True, help="The run ID to filter by")
    sub.add_argument(
        "-s", "--stream", action="store_true", help="Stream events as they're ready"
    )
    sub.add_argument(
        "-o",
        "--offset",
        type=int,
        help="A cursor, allows you to offset into the stream. Negative indexes allow you to index from the current end of the stream.",
    )
    sub.add_argument("-l", "--limit", type=int, help="How many records to return.")
    sub.add_argument(
        "-t",
        "--timeout",
        help="An optional amount of time to block for new events. If the timeout expires, an empty list will be returned.",
        type=float,
    )
    sub.set_defaults(func=Event.list)

    # Files
    sub = subparsers.add_parser("files.create")
    sub.add_argument("--purpose", default="plan")
    sub.add_argument("-p", "--path", required=True)
    sub.set_defaults(func=File.create)

    sub = subparsers.add_parser("files.get")
    sub.add_argument("-i", "--id", required=True, help="The file ID")
    sub.set_defaults(func=File.get)

    sub = subparsers.add_parser("files.list")
    sub.set_defaults(func=File.list)

    # Plans
    sub = subparsers.add_parser("plans.create")
    sub.add_argument(
        "-d", "--description", help="A human-readable description of this plan"
    )
    sub.add_argument(
        "-f", "--file", help="An uploaded JSONL file corresponding to the plan"
    )
    sub.add_argument("-p", "--path")
    sub.set_defaults(func=Plan.create)

    sub = subparsers.add_parser("plans.get")
    sub.add_argument("-i", "--id", required=True, help="The plan ID")
    sub.set_defaults(func=Plan.get)

    sub = subparsers.add_parser("plans.delete")
    sub.add_argument("-i", "--id", required=True, help="The plan ID")
    sub.set_defaults(func=Plan.delete)

    sub = subparsers.add_parser("plans.list")
    sub.set_defaults(func=Plan.list)

    # Runs
    sub = subparsers.add_parser("runs.create")
    sub.add_argument(
        "-m",
        "--model",
        required=True,
        help="A model (most commonly a snapshot ID) to start the run from",
    )
    sub.add_argument("-p", "--plan", required=True, help="The plan to apply.")
    sub.set_defaults(func=Run.create)

    sub = subparsers.add_parser("runs.get")
    sub.add_argument("-i", "--id", required=True, help="The run ID")
    sub.set_defaults(func=Run.get)

    sub = subparsers.add_parser("runs.update")
    sub.add_argument("-i", "--id", required=True, help="The run ID")
    sub.set_defaults(func=Run.get)

    sub = subparsers.add_parser("runs.list")
    sub.set_defaults(func=Run.list)

    ## Snapshots
    sub = subparsers.add_parser("snapshots.list")
    sub.set_defaults(func=Snapshot.list)

    sub = subparsers.add_parser("snapshots.create")
    sub.add_argument(
        "-e", "--engine", required=True, help="The engine this snapshot is running on"
    )
    sub.add_argument(
        "-b", "--branch", required=True, help="The branch to turn into a snapshot"
    )
    sub.add_argument(
        "-d", "--description", help="A human-readable description of this snapshot"
    )
    sub.add_argument(
        "-t",
        "--timeout",
        help="An optional amount of time to block for the snapshot to transition from pending. If the timeout expires, a pending snapshot will be returned.",
        type=float,
    )
    sub.set_defaults(func=Snapshot.create)

    sub = subparsers.add_parser("snapshots.get")
    sub.add_argument("-e", "--engine", help="The engine this snapshot is running on")
    sub.add_argument("-i", "--id", required=True, help="The snapshot ID")
    sub.add_argument(
        "-t",
        "--timeout",
        help="An optional amount of time to block for the snapshot to transition from pending. If the timeout expires, a pending snapshot will be returned.",
        type=float,
    )
    sub.set_defaults(func=Snapshot.get)

    sub = subparsers.add_parser("snapshots.delete")
    sub.add_argument("-i", "--id", required=True, help="The snapshot ID")
    sub.set_defaults(func=Snapshot.delete)

    # Tags
    sub = subparsers.add_parser("tags.create")
    sub.add_argument("-i", "--id", help="The ID of the tag to create", required=True)
    sub.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Whether to overwrite an existing tag with this ID",
    )
    sub.add_argument(
        "-m",
        "--model",
        help="Which model (most commonly a snapshot ID) to tag",
        required=True,
    )
    sub.set_defaults(func=Tag.create)

    sub = subparsers.add_parser("tags.get")
    sub.add_argument("-i", "--id", required=True, help="The tag ID")
    sub.set_defaults(func=Tag.get)

    sub = subparsers.add_parser("tags.delete")
    sub.add_argument("-i", "--id", required=True, help="The tag ID")
    sub.set_defaults(func=Tag.delete)

    sub = subparsers.add_parser("tags.list")
    sub.set_defaults(func=Tag.list)

    ## Updates
    sub = subparsers.add_parser("updates.create")
    sub.add_argument(
        "-e", "--engine", required=True, help="The engine this update is running on"
    )
    sub.add_argument(
        "-b", "--branch", required=True, help="The branch to make a update on"
    )
    sub.add_argument(
        "-c",
        "--context",
        help="The data to train on: either a string, a list of strings, or a list of tokens (encoded as JSON)",
        required=True,
    )
    sub.add_argument(
        "--mask",
        help="An array of numbers specifying relative training weight for each token (encoded as JSON)",
    )
    sub.add_argument("-s", "--scale", help="How large of a update to take", type=float)
    sub.add_argument(
        "-t",
        "--timeout",
        help="An optional amount of time to block for the update to transition from pending. If the timeout expires, a pending update will be returned.",
        type=float,
    )
    sub.set_defaults(func=Update.create)

    sub = subparsers.add_parser("updates.get")
    sub.add_argument(
        "-e", "--engine", required=True, help="The engine this update is running on"
    )
    sub.add_argument("-i", "--id", required=True, help="The update ID")
    sub.add_argument(
        "-t",
        "--timeout",
        help="An optional amount of time to block for the update to transition from pending. If the timeout expires, a pending update will be returned.",
        type=float,
    )
    sub.set_defaults(func=Update.get)



import json
import tempfile
import os.path as osp
from openai.finetune import data_streams
import data_gym
import blobfile


def val_stream_from_files():
    text = ",".join(str(i) for i in range(1000))
    tmpdir = tempfile.mkdtemp()
    fname = osp.join(tmpdir, "a.jsonl")
    print(fname)
    enc = data_gym.text.get_encoding("reversible_50000")
    with open(fname, "w") as fh:
        fh.write(json.dumps({"data": text}))
        fh.write("\n")
    stream = data_streams.stream_from_files(
        [fname], n_ctx=17, seed=0, shuffle=False, overlap=0
    )
    roundtrip = "".join([enc.decode(chunk["tokens"][0].tolist()) for chunk in stream])
    # pylint: disable=bad-str-strip-call
    assert roundtrip.rstrip("<|endoftext|>") == text


def val_estimate():
    paths = blobfile.glob("gs://easyfinetune/reddit-cleanjokes-*.jsonl")
    for path in paths:
        size_sub, stderr_sub = data_streams.get_num_tokens([path], n_samples=100)
        size_all, stderr_all = data_streams.get_num_tokens([path], n_samples=int(1e6))
        assert stderr_all == 0
        z = (size_all - size_sub) / stderr_sub
        errpct = 100 * (size_all - size_sub) / size_all
        assert abs(z) < 3
        print(
            f"Est: {size_sub:.1f} ± {stderr_sub:.1f}. True: {size_all}.",
            f"%err={errpct:.1f}, z={z:.1f}",
        )



import os

from setuptools import find_packages, setup

version_contents = {}
with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "openai/version.py")
) as f:
    exec(f.read(), version_contents)

setup(
    name="openai",
    description="Python client library for the OpenAI API",
    version=version_contents["VERSION"],
    install_requires=[
        'requests >= 2.20; python_version >= "3.0"',
        'requests[security] >= 2.20; python_version < "3.0"',
    ],
    extras_require={"fine-tune": ["numpy", "transformers"]},
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    scripts=["bin/openai"],
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"openai": ["data/ca-certificates.crt"]},
    author="OpenAI",
    author_email="support@openai.com",
    url="https://github.com/openai/openai-python",
)



#!/usr/bin/env python
import argparse
from openai.finetune import data_streams
import data_gym
import time
from termcolor import colored


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("train_url")
    parser.add_argument("--seed", type=int)
    args = parser.parse_args()
    streamer = data_streams.stream_from_files(
        files=[args.train_url], n_ctx=100, seed=args.seed, overlap=0
    )
    enc = data_gym.text.get_encoding("reversible_50000")
    for x in streamer:
        tokens = x["tokens"][0].tolist()
        weights = x["weights"][0].tolist()
        assert len(tokens) == len(weights)
        for (t, w) in zip(tokens, weights):
            color = "white" if w > 0.1 else "blue"
            s = enc.decode([t])
            if s == "<|endoftext|>":
                print(colored("□", color="magenta"))
            else:
                print(colored(s, color=color), end="", flush=True)
            time.sleep(0.1)


if __name__ == "__main__":
    main()



from __future__ import absolute_import, division, print_function

from openai.api_resources.abstract import (
    APIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
)


class Plan(
    CreateableAPIResource,
    UpdateableAPIResource,
    ListableAPIResource,
    DeletableAPIResource,
):
    OBJECT_NAME = "plan"



from __future__ import absolute_import, division, print_function

import json
import os
import tempfile

from openai import util
from openai.api_resources.abstract import (
    APIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
)
from openai.openai_object import OpenAIObject
from openai.six.moves.urllib.parse import quote_plus
from openai.util import log_info


class Event(ListableAPIResource):
    OBJECT_NAME = "event"

    @classmethod
    def list(cls, stream=False, **params):
        if stream:
            return cls._events_stream(**params)
        return super().list(**params)

    @classmethod
    def _events_stream(cls, **params):
        while True:
            resp = cls.list(**params)
            for item in resp.data:
                yield item
            if not resp.has_more:
                return
            elif len(resp.data) > 0:
                params["offset"] = resp.data[-1].data.created_by.lineno + 1
            if params.get("limit") is None:
                params["limit"] = 100
            if params.get("timeout") is None:
                params["timeout"] = 10



from openai import encoding, util
from openai.api_resources.abstract import (
    APIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    ListableAPIResource,
    UpdateableAPIResource,
)
from openai.six.moves.urllib.parse import quote_plus


class CompletionConfig(
    CreateableAPIResource, ListableAPIResource, DeletableAPIResource
):
    OBJECT_NAME = "experimental.completion_config"



from __future__ import absolute_import, division, print_function

from openai import api_requestor, six, util
from openai.openai_object import OpenAIObject

from openai.six.moves.urllib.parse import quote_plus


class ListObject(OpenAIObject):
    OBJECT_NAME = "list"

    def list(self, api_key=None, openai_version=None, openai_account=None, **params):
        openai_object = self._request(
            "get",
            self.get("url"),
            api_key=api_key,
            openai_version=openai_version,
            openai_account=openai_account,
            **params
        )
        openai_object._retrieve_params = params
        return openai_object

    def create(
        self,
        api_key=None,
        idempotency_key=None,
        openai_version=None,
        openai_account=None,
        **params
    ):
        return self._request(
            "post",
            self.get("url"),
            api_key=api_key,
            idempotency_key=idempotency_key,
            openai_version=openai_version,
            openai_account=openai_account,
            **params
        )

    def retrieve(
        self, id, api_key=None, openai_version=None, openai_account=None, **params
    ):
        url = "%s/%s" % (self.get("url"), quote_plus(util.utf8(id)))
        return self._request(
            "get",
            url,
            api_key=api_key,
            openai_version=openai_version,
            openai_account=openai_account,
            **params
        )

    def _request(
        self,
        method_,
        url_,
        api_key=None,
        idempotency_key=None,
        openai_version=None,
        openai_account=None,
        **params
    ):
        api_key = api_key or self.api_key
        openai_version = openai_version or self.openai_version
        openai_account = openai_account or self.openai_account

        requestor = api_requestor.APIRequestor(
            api_key, api_version=openai_version, account=openai_account
        )
        headers = util.populate_headers(idempotency_key)
        response, api_key = requestor.request(method_, url_, params, headers)
        openai_object = util.convert_to_openai_object(
            response, api_key, openai_version, openai_account
        )
        return openai_object

    def __getitem__(self, k):
        if isinstance(k, six.string_types):
            return super(ListObject, self).__getitem__(k)
        else:
            raise KeyError(
                "You tried to access the %s index, but ListObject types only "
                "support string keys. (HINT: List calls return an object with "
                "a 'data' (which is the data array). You likely want to call "
                ".data[%s])" % (repr(k), repr(k))
            )

    def __iter__(self):
        return getattr(self, "data", []).__iter__()

    def __len__(self):
        return getattr(self, "data", []).__len__()

    def __reversed__(self):
        return getattr(self, "data", []).__reversed__()

    def auto_paging_iter(self):
        page = self

        while True:
            if (
                "ending_before" in self._retrieve_params
                and "starting_after" not in self._retrieve_params
            ):
                for item in reversed(page):
                    yield item
                page = page.previous_page()
            else:
                for item in page:
                    yield item
                page = page.next_page()

            if page.is_empty:
                break

    @classmethod
    def empty_list(cls, api_key=None, openai_version=None, openai_account=None):
        return cls.construct_from(
            {"data": []},
            key=api_key,
            openai_version=openai_version,
            openai_account=openai_account,
            last_response=None,
        )

    @property
    def is_empty(self):
        return not self.data

    def next_page(
        self, api_key=None, openai_version=None, openai_account=None, **params
    ):
        if not self.has_more:
            return self.empty_list(
                api_key=api_key,
                openai_version=openai_version,
                openai_account=openai_account,
            )

        last_id = self.data[-1].id

        params_with_filters = self._retrieve_params.copy()
        params_with_filters.update({"starting_after": last_id})
        params_with_filters.update(params)

        return self.list(
            api_key=api_key,
            openai_version=openai_version,
            openai_account=openai_account,
            **params_with_filters
        )

    def previous_page(
        self, api_key=None, openai_version=None, openai_account=None, **params
    ):
        if not self.has_more:
            return self.empty_list(
                api_key=api_key,
                openai_version=openai_version,
                openai_account=openai_account,
            )

        first_id = self.data[0].id

        params_with_filters = self._retrieve_params.copy()
        params_with_filters.update({"ending_before": first_id})
        params_with_filters.update(params)

        return self.list(
            api_key=api_key,
            openai_version=openai_version,
            openai_account=openai_account,
            **params_with_filters
        )
